﻿using System;
using System.Collections.ObjectModel;
using System.Drawing;
using System.Xml;
using System.Globalization;
using System.Net;
using System.IO;
using System.Windows.Forms;
using System.Collections.Generic;
using ThinkGeo.MapSuite.Core;

namespace ValueStyleWithCountries
{
    public partial class TestForm : Form
    {
        private MapEngine mapEngine = new MapEngine();
        private Bitmap bitmap = null;
        public int i;
        public ValueStyle styles = new ValueStyle();
        public string iconString;
        public bool first = true;
        public Collection<int> ids = new Collection<int>();
        public string prevString;
        public GeoImage prevImage = new GeoImage();
       
       public TestForm()
        {
            InitializeComponent();
        }

        private void TestForm_Load(object sender, EventArgs e)
        {
            mapEngine.CurrentExtent = ExtentHelper.GetDrawingExtent(new RectangleShape(-160, 60, 160, -60),  Map.Width, Map.Height);
            mapEngine.BackgroundFillBrush = new GeoSolidBrush(GeoColor.StandardColors.Transparent);
            
            GetStuff();
        }

        private void GetStuff()
        {
            styles.ColumnName = "name";
            getAllData(@"fcstWeather.kml");
            
            DrawImage();
        }

        private void getAllData(string p)
        {
            XmlDocument kml = new XmlDocument();
            kml.Load(p);

            XmlNamespaceManager manager = new XmlNamespaceManager(kml.NameTable);
            manager.AddNamespace("kk", kml.DocumentElement.NamespaceURI);

            if (kml.DocumentElement.HasChildNodes)
            {
                XmlNodeList styleList = kml.DocumentElement.ChildNodes[0].SelectNodes("kk:Style", manager);
                foreach (XmlNode item in styleList)
                {
                    string styleId = item.Attributes["id"].Value;
                        
                    ValueItem style = new ValueItem();
                    style.Value = styleId;
                    if (item.SelectSingleNode("kk:IconStyle", manager) != null)
                    {
                        XmlNode iconNode = item.SelectSingleNode("kk:IconStyle", manager);
                        XmlNode hrefNode = iconNode.SelectSingleNode("kk:Icon", manager);
                        iconString = hrefNode.SelectSingleNode("kk:href", manager).InnerText.ToString();
                        PointStyle points = new PointStyle();
                        points.PointType = PointType.Bitmap;
                        if (iconString != prevString)
                        {
                            points.Image = GetImage(iconString);
                            prevString = iconString;
                        }

                        else if (iconString == prevString)
                        {
                            points.Image = prevImage;
                        }

                        try
                        {
                            Single heading = Convert.ToSingle(iconNode.SelectSingleNode("kk:heading", manager).InnerText.ToString());
                            points.RotationAngle = heading;
                        }

                        catch (System.NullReferenceException)
                        {
                        }

                        try
                        {
                            double scale = Convert.ToDouble(iconNode.SelectSingleNode("kk:scale", manager).InnerText);
                            points.ImageScale = scale;
                        }

                        catch (System.NullReferenceException)
                        {
                        }
                        style.DefaultPointStyle = points;
                        styles.ValueItems.Add(style);
                    }

                    if (item.SelectSingleNode("kk:LineStyle", manager) != null)
                    {
                        LineStyle lines = new LineStyle();
                        GeoPen pen = new GeoPen();
                        GeoColor color = new GeoColor();
                        Int32 width = new Int32();
                        try
                        {
                            XmlNode colorxml = item.SelectSingleNode("kk:LineStyle", manager);
                            XmlNode colornode = colorxml.SelectSingleNode("kk:color", manager);
                            XmlNode widthnode = colorxml.SelectSingleNode("kk:width", manager);
                            string color2 = colornode.InnerText.ToString();
                            width = Int32.Parse(widthnode.InnerText);
                            color = GeoColor.FromWin32(Int32.Parse(color2, NumberStyles.HexNumber));
                        }

                        catch (System.NullReferenceException)
                        {
                        }

                        pen.Color = color;
                        pen.Width = width;
                        lines.CenterPen = pen;
                        style.DefaultLineStyle = lines;
                        styles.ValueItems.Add(style);
                    }

                    if (item.SelectSingleNode("kk:PolyStyle", manager) != null)
                    {
                        AreaStyle poly = new AreaStyle();
                        GeoSolidBrush brush = new GeoSolidBrush();
                        GeoColor color = new GeoColor();
                        try
                        {
                            XmlNode polyxml = item.SelectSingleNode("kk:PolyStyle", manager);
                            XmlNode colorxml = polyxml.SelectSingleNode("kk:fill", manager);
                            XmlNode colornode = polyxml.SelectSingleNode("kk:color", manager);
                            string color2 = colornode.InnerText.ToString();
                            int nA = int.Parse(color2.Substring(0, 2), NumberStyles.HexNumber);
                            int nR = int.Parse(color2.Substring(2, 2), NumberStyles.HexNumber);
                            int nG = int.Parse(color2.Substring(4, 2), NumberStyles.HexNumber);
                            int nB = int.Parse(color2.Substring(6, 2), NumberStyles.HexNumber);
                            color = GeoColor.FromArgb(nA, nR, nG, nB);
                        }

                        catch (System.NullReferenceException)
                        {
                        }

                        brush.Color = color;
                        style.DefaultAreaStyle.FillSolidBrush = brush;
                        styles.ValueItems.Add(style);
                    }
                }
            }

            treeView1.BeginUpdate();
            treeView1.CheckBoxes = true;
            TreeNode node = new TreeNode();
            node = treeView1.Nodes.Add("Folder");
            node.Checked = true;
            treeView1.EndUpdate();

            XmlNodeList folders = kml.DocumentElement.ChildNodes[0].SelectNodes("kk:Folder", manager);
            processFolders(manager, folders, node);
        }
        
        void processFolders(XmlNamespaceManager manager, XmlNodeList folders, TreeNode node)
        {
            foreach (XmlNode folder in folders)
            {
                XmlNodeList subfolders = folder.SelectNodes("kk:Folder", manager);

                TreeNode node2 = new TreeNode(folder.Name);
                node2.Name = folder.Name;
                node2.Text = folder.SelectSingleNode("kk:name", manager).InnerText;

                if (first == true)
                {
                    first = false;
                }

                node.Nodes.Add(node2);

                if (subfolders.Count != 0)
                {
                    processFolders(manager, subfolders, node2);
                }

                else
                {
                    InMemoryFeatureLayer layer = new InMemoryFeatureLayer();
                    layer.Name = folder.SelectSingleNode("kk:name", manager).InnerText;
                    ProcessPlacemark(folder, manager, layer);
                    layer.ZoomLevelSet.ZoomLevel01.CustomStyles.Add(styles);
                    layer.ZoomLevelSet.ZoomLevel01.ApplyUntilZoomLevel = ApplyUntilZoomLevel.Level20;
                    mapEngine.StaticLayers.Add(layer);
                }
            }
        }

        private void ProcessPlacemark(XmlNode placemark, XmlNamespaceManager manager, InMemoryFeatureLayer layer)
        {
            Collection<Feature> tempCollection = new Collection<Feature>();


            XmlNodeList placeMarkXML = placemark.SelectNodes("kk:Placemark", manager);

            /*
                * Point --> Done
                * LineString --> Done
                * LinearRing --> Done
                * Polygon
                * MultiGeometry --> Done
                * Model 
                */

            foreach (XmlNode mark in placeMarkXML)
            {
                XmlNode stylesNode = mark.SelectSingleNode("kk:styleUrl", manager);
                Dictionary<string, string> columnValues = new Dictionary<string, string>();
                /*
                XmlNode descriptionNode = placemark.SelectSingleNode("kk:description", manager);
                if (descriptionNode != null)
                {
                    columnValues.Add("popupHTML", descriptionNode.InnerXml);
                }
                */
                processNodes(mark, manager, tempCollection, columnValues, stylesNode, layer);

                XmlNodeList multigeometry = mark.SelectNodes("kk:MultiGeometry", manager);
                foreach (XmlNode geometry in multigeometry)
                {
                    processNodes(geometry, manager, tempCollection, columnValues, stylesNode, layer);
                }

                XmlNodeList polygon = mark.SelectNodes("kk:Polygon", manager);
                foreach (XmlNode geometry in polygon)
                {
                    //processNodes(geometry, manager, tempCollection, columnValues, stylesNode, fd);
                    XmlNodeList outer = geometry.SelectNodes("kk:outerBoundaryIs", manager);
                    foreach (XmlNode line in outer)
                    {
                        processNodes(line, manager, tempCollection, columnValues, stylesNode, layer);
                    }

                    XmlNodeList inner = geometry.SelectNodes("kk:innerBoundaryIs", manager);
                    foreach (XmlNode line in inner)
                    {
                        processNodes(line, manager, tempCollection, columnValues, stylesNode, layer);
                    }
                }
            }
        }

        void processNodes(XmlNode geometry, XmlNamespaceManager manager, Collection<Feature> tempCollection, Dictionary<string, string> columnValues, XmlNode stylesNode, InMemoryFeatureLayer layer)
        {
            //XmlNodeList polygonList = geometry.SelectNodes("kk:Polygon", manager);
            //foreach (XmlNode node in polygonList)
            //{

            //}

            XmlNodeList linearRingList = geometry.SelectNodes("kk:LinearRing", manager);
            foreach (XmlNode node in linearRingList)
            {
                string value = node.SelectSingleNode("kk:coordinates", manager).InnerText.Trim();
                //TODO: the document says use " ", not "\r\n"
                string[] values = value.Split(new string[] { " ", "\r\n" }, StringSplitOptions.RemoveEmptyEntries);
                Collection<Vertex> Vertices = new Collection<Vertex>();
                foreach (string item in values)
                {
                    string[] items = item.Split(new char[] { ',' }, StringSplitOptions.RemoveEmptyEntries);
                    Vertex vertex = new Vertex(double.Parse(items[0]), double.Parse(items[1]));
                    Vertices.Add(vertex);
                }
                RingShape ringShape = new RingShape(Vertices);
                if (columnValues.Count == 0 && stylesNode != null)
                {
                    string styleUrl = stylesNode.InnerText;
                    if (styleUrl.StartsWith("#"))
                    {
                        try
                        {
                            string name = styleUrl.TrimStart('#').ToString();
                            columnValues.Add("name", name);
                        }

                        catch (KeyNotFoundException)
                        {
                        }
                    }

                    else
                    {
                        try
                        {
                            string name = styleUrl.ToString();
                            columnValues.Add("name", name);
                        }

                        catch (KeyNotFoundException)
                        {
                        }
                    }
                }
                //tempFeatures.Add(new Feature(ringShape, columnValues));
                tempCollection.Add(new Feature(ringShape, columnValues));
            }

            XmlNodeList pointList = geometry.SelectNodes("kk:Point", manager);
            foreach (XmlNode item in pointList)
            {
                string pointValue = item.SelectSingleNode("kk:coordinates", manager).InnerText.Trim();
                string[] pointValues = pointValue.Split(new char[] { ',' }, StringSplitOptions.RemoveEmptyEntries);
                PointShape point = new PointShape(double.Parse(pointValues[0]), double.Parse(pointValues[1]));
                if (columnValues.Count == 0 && stylesNode != null)
                {
                    string styleUrl = stylesNode.InnerText;
                    if (styleUrl.StartsWith("#"))
                    {
                        try
                        {
                            string name = styleUrl.TrimStart('#').ToString();
                            columnValues.Add("name", name);
                        }

                        catch (KeyNotFoundException)
                        {
                        }
                    }
                }
                //tempFeatures.Add(new Feature(point, columnValues));
                tempCollection.Add(new Feature(point, columnValues));
            }

            XmlNodeList lineStringList = geometry.SelectNodes("kk:LineString", manager);
            foreach (XmlNode node in lineStringList)
            {
                string value = node.SelectSingleNode("kk:coordinates", manager).InnerText.Trim();
                //TODO: the document says use " ", not "\r\n"
                string[] values = value.Split(new string[] { " ", "\r\n" }, StringSplitOptions.RemoveEmptyEntries);
                Collection<Vertex> Vertices = new Collection<Vertex>();
                foreach (string item in values)
                {
                    string[] items = item.Split(new char[] { ',' }, StringSplitOptions.RemoveEmptyEntries);
                    Vertex vertex = new Vertex(double.Parse(items[0]), double.Parse(items[1]));
                    Vertices.Add(vertex);
                }
                LineShape lineShape = new LineShape(Vertices);

                if (columnValues.Count == 0 && stylesNode != null)
                {
                    string styleUrl = stylesNode.InnerText;
                    if (styleUrl.StartsWith("#"))
                    {
                        try
                        {
                            string name = styleUrl.TrimStart('#').ToString();
                            columnValues.Add("name", name);
                        }

                        catch (KeyNotFoundException)
                        {

                        }
                    }
                }
                //tempFeatures.Add(new Feature(lineShape, columnValues));
                tempCollection.Add(new Feature(lineShape, columnValues));
            }
            foreach (Feature item in tempCollection)
            {
                WellKnownType wellKnownType = item.GetWellKnownType();
                if (wellKnownType == WellKnownType.Polygon || wellKnownType == WellKnownType.Multipolygon ||
                    wellKnownType == WellKnownType.Line || wellKnownType == WellKnownType.Multiline ||
                    wellKnownType == WellKnownType.Point || wellKnownType == WellKnownType.Multipoint)
                {
                    layer.InternalFeatures.Add(item);
                }

            }
            tempCollection.Clear();
        }

        private GeoImage GetImage(string sURL)
        {
            MemoryStream ms = new MemoryStream(new WebClient().DownloadData(sURL));
            prevImage = new GeoImage(ms);
            return new GeoImage(ms);
        }

        private void DrawImage()
        {
            if (bitmap != null) 
            { 
                bitmap.Dispose(); 
            }

            bitmap = new Bitmap(Map.Width, Map.Height);
            mapEngine.OpenAllLayers();

            mapEngine.DrawStaticLayers(bitmap, GeographyUnit.DecimalDegree);
            
            mapEngine.DrawDynamicLayers(bitmap, GeographyUnit.DecimalDegree);
            mapEngine.CloseAllLayers();

            Map.Image = bitmap;
        }

        private void ToolBar_ButtonClick(object sender, ToolBarButtonClickEventArgs e)
        {
            switch (e.Button.Tag.ToString())
            {
                case "Zoom In":
                    mapEngine.CurrentExtent.ScaleDown(50);
                    break;
                case "Zoom Out":
                    mapEngine.CurrentExtent.ScaleUp(50);
                    break;
                case "Full Extent":
                    mapEngine.CurrentExtent = ExtentHelper.GetDrawingExtent(new RectangleShape(-180.0, 83.0, 180.0, -90.0), Map.Width, Map.Height);
                    break;
                case "Pan Left":
                    mapEngine.CurrentExtent = ExtentHelper.Pan(mapEngine.CurrentExtent, PanDirection.Left, 20);
                    break;
                case "Pan Right":
                    mapEngine.CurrentExtent = ExtentHelper.Pan(mapEngine.CurrentExtent, PanDirection.Right, 20);
                    break;
                case "Pan Up":
                    mapEngine.CurrentExtent = ExtentHelper.Pan(mapEngine.CurrentExtent, PanDirection.Up, 20);
                    break;
                case "Pan Down":
                    mapEngine.CurrentExtent = ExtentHelper.Pan(mapEngine.CurrentExtent, PanDirection.Down, 20);
                    break;
                default:
                    break;
            }
            DrawImage();
        }

        private void btnClose_Click(object sender, EventArgs e)
        {
            this.Close();
        }

        private void Map_MouseMove(object sender, MouseEventArgs e)
        {
            //Displays the X and Y in screen coordinates.
            statusStrip1.Items["toolStripStatusLabelScreen"].Text = "X:" + e.X + " Y:" + e.Y;

            //Gets the PointShape in world coordinates from screen coordinates.
            PointShape pointShape = ExtentHelper.ToWorldCoordinate(mapEngine.CurrentExtent, new ScreenPointF(e.X, e.Y), Map.Width, Map.Height);

            //Displays world coordinates.
            statusStrip1.Items["toolStripStatusLabelWorld"].Text = "(world) X:" + Math.Round(pointShape.X, 4) + " Y:" + Math.Round(pointShape.Y, 4);
        }

        private void button1_Click(object sender, EventArgs e)
        {
            textBox1.Text = ids.Count.ToString();
        }

        private void treeView1_AfterCheck(object sender, TreeViewEventArgs e)
        {
            if (e.Action != TreeViewAction.Unknown)
            {
                if (e.Node.Checked == true)
                    ids.Add(Convert.ToInt32(e.Node.Name));
                else if (e.Node.Checked == false)
                    ids.Remove(Convert.ToInt32(e.Node.Name));
                
                if (e.Node.Nodes.Count > 0)
                {
                    /* Calls the CheckAllChildNodes method, passing in the current 
                    Checked value of the TreeNode whose checked state changed. */
                    this.CheckAllChildNodes(e.Node, e.Node.Checked);
                }

                DrawImage();
            }
        }

        private void CheckAllChildNodes(TreeNode treeNode, bool nodeChecked)
        {
            foreach (TreeNode node in treeNode.Nodes)
            {
                if (node.Checked == false)
                    ids.Add(Convert.ToInt32(node.Name));
                else if (node.Checked == true)
                    ids.Remove(Convert.ToInt32(node.Name));
                node.Checked = nodeChecked;
                if (node.Nodes.Count > 0)
                {
                    // If the current node has child nodes, call the CheckAllChildsNodes method recursively. 
                    this.CheckAllChildNodes(node, nodeChecked);
                }
            }
        }
    }
}
